'''
TensorFlowでキルミーアイコン686枚によるキルミー的アニメ絵分類
https://qiita.com/domkade/items/fc9903c2119fdfa9a234

独自のデータを使用してAlexNetを実装する。
'''

import os
import numpy as np
import cv2
import tensorflow as tf
import tf_make_dataset
import random


def init_weight_variable(shape):
    """
    重みを正規乱数で初期化する関数
    """
    initial = tf.truncated_normal(shape, stddev=0.1)
    return tf.Variable(initial)


def init_bias_variable(shape):
    """
    バイアスを定数(今回は、0.1)で初期化する関数
    """
    initial = tf.constant(0.1, shape=shape)
    return tf.Variable(initial)


def conv2d(x, W):
    """
    畳み込み演算を行うため関数
    """
    return tf.nn.conv2d(x, W, strides=[1, 1, 1, 1], padding='SAME')


def max_pool_3x3(x):
    """
    3ｘ3のマックスプーリングを行うため関数
    """
    return tf.nn.max_pool(x, ksize=[1, 3, 3, 1], strides=[1, 2, 2, 1], padding='SAME')


img_dir = "./image"
IMG_SIZE = 224
COLOR_CHANNELS = 3  # RGB
IMG_PIXELS = IMG_SIZE * IMG_SIZE * COLOR_CHANNELS

TRAIN_NUM = 40
TEST_NUM = 10
CLASS_NUM = 3

train_image, train_label, test_image, test_label = tf_make_dataset.make(
    img_dir, img_size=IMG_SIZE, train_num=TRAIN_NUM, test_num=TEST_NUM)

# x:入力値,y_:出力値のplaceholderセット
x = tf.placeholder(tf.float32, shape=[None, IMG_PIXELS])
y_ = tf.placeholder(tf.float32, shape=[None, CLASS_NUM])

# [バッチ数、縦、横、チャンネル数]
x_image = tf.reshape(x, [-1, IMG_SIZE, IMG_SIZE, COLOR_CHANNELS])

# 畳み込み層1（フィルター数は96個）
# フィルターのパラメタをセット
# [縦、横、チャンネル数、フィルター数]
W_conv1 = init_weight_variable([11, 11, COLOR_CHANNELS, 96])
# 96個のバイアスをセット
b_conv1 = init_bias_variable([96])
# 畳み込み演算後に、ReLU関数適用
h_conv1 = tf.nn.relu(tf.nn.conv2d(x_image, W_conv1, strides=[
                     1, 4, 4, 1], padding='SAME') + b_conv1)
# 3x3のMAXプーリングを実行
h_pool1 = max_pool_3x3(h_conv1)

# 畳み込み層2（フィルター数は256個）
# フィルターのパラメタをセット
# チャンネル数が96なのは、畳み込み層１のフィルター数が96だから。
# 96個フィルターがあると、出力結果が[-1, IMG_SIZE, IMG_SIZE, 96]というshapeになる。
# 入力のチャンネル数と重みのチャンネル数を合わせる。
W_conv2 = init_weight_variable([5, 5, 96, 256])
b_conv2 = init_bias_variable([256])
h_conv2 = tf.nn.relu(conv2d(h_pool1, W_conv2) + b_conv2)
h_pool2 = max_pool_3x3(h_conv2)

# 畳み込み層3（フィルター数は384個）
W_conv3 = init_weight_variable([3, 3, 256, 384])
b_conv3 = init_bias_variable([384])
h_conv3 = tf.nn.relu(conv2d(h_pool2, W_conv3) + b_conv3)

# 畳み込み層4（フィルター数は384個）
W_conv4 = init_weight_variable([3, 3, 384, 384])
b_conv4 = init_bias_variable([384])
h_conv4 = tf.nn.relu(conv2d(h_conv3, W_conv4) + b_conv4)

# 畳み込み層5（フィルター数は256個）
W_conv5 = init_weight_variable([3, 3, 384, 256])
b_conv5 = init_bias_variable([256])
h_conv5 = tf.nn.relu(conv2d(h_conv4, W_conv5) + b_conv5)
h_pool3 = max_pool_3x3(h_conv5)

# 全結合層1（ノードの数は4096個）
W_fc1 = init_weight_variable([7*7 * 256, 4096])
b_fc1 = init_bias_variable([4096])
# 全結合層の入力仕様に合わせて、2次元にreshape
h_pool3_flat = tf.reshape(h_pool3, [-1, 7*7*256])
h_fc1 = tf.nn.relu(tf.matmul(h_pool3_flat, W_fc1) + b_fc1)

# ドロップアウト処理
# keep_probは、ドロップアウトさせない率
keep_prob = tf.placeholder(tf.float32)
h_fc1_drop = tf.nn.dropout(h_fc1, keep_prob)

# 全結合層2（ノードの数は4096個）
W_fc2 = init_weight_variable([4096, 4096])
b_fc2 = init_bias_variable([4096])
h_fc2 = tf.nn.relu(tf.matmul(h_fc1_drop, W_fc2) + b_fc2)
h_fc2_drop = tf.nn.dropout(h_fc2, keep_prob)

# 出力層
W_fc3 = init_weight_variable([4096, CLASS_NUM])
b_fc3 = init_bias_variable([CLASS_NUM])
y_conv = tf.matmul(h_fc2_drop, W_fc3) + b_fc3

# softmaxと交差エントロピーを合わせてやってくれるやつっぽい。あとで確認する。
cross_entropy = tf.reduce_mean(
    tf.nn.softmax_cross_entropy_with_logits_v2(logits=y_conv, labels=y_))
# AdamOptimizerの学習率を0.0001に設定している。デフォルトが0.001だが、
# 複雑なモデルになるとより小さくしないと発散してうまくいかないケースがあるらしい。
train_step = tf.train.AdamOptimizer(1e-4).minimize(cross_entropy)
correct_prediction = tf.equal(tf.argmax(y_conv, 1), tf.argmax(y_, 1))
accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))

# InteractiveSessoin作成
sess = tf.InteractiveSession()
sess.run(tf.global_variables_initializer())

EPOCHS = 50  # 学習ステップ数
BATCH_SIZE = 10  # バッチサイズ

for i in range(EPOCHS):
    random_seq = list(range(len(train_image)))
    random.shuffle(random_seq)
    for j in range(int(len(train_image)/BATCH_SIZE)):
        batch = BATCH_SIZE * j
        train_image_batch = []
        train_label_batch = []
        for k in range(BATCH_SIZE):
            train_image_batch.append(train_image[random_seq[batch + k]])
            train_label_batch.append(train_label[random_seq[batch + k]])

        train_step.run(feed_dict={x: train_image_batch,
                                  y_: train_label_batch, keep_prob: 0.5})

    # 毎ステップ、学習データに対する正答率を表示
    train_accuracy = accuracy.eval(feed_dict={
        x: train_image, y_: train_label, keep_prob: 1.0})
    print("%d epoch, training accuracy %g" % (i + 1, train_accuracy))

test_accuracy = accuracy.eval(feed_dict={
    x: test_image, y_: test_label, keep_prob: 1.0})
print("test accuracy %g" % (test_accuracy))
